import argparse
import logging
import random
import re
import sys
import unittest

from typing import Callable

from alembic.config import Config as AlembicConfig
from alembic.script import ScriptDirectory
from alembic.migration import MigrationContext
from attrs import define
from sqlalchemy import MetaData

from pyvoog.app import Application
from pyvoog.db import setup_database
from pyvoog.logging import setup_logging
from pyvoog.testing.signals import app_ctx_pushed

@define
class TestRunner:

    """ A high-level test runner class for quick bootstrapping of a test
    runner script. Provides argument handling (e.g. `--filter` and `--
    verbose`; use `--help` for a synopsis), test case filtering and
    discovery, running and reporting results from test suites.

    Attributes, all optional:

    - alembic_config_fn - Alembic configuration file name. If passed, the
      database is checked to be fully migrated.
    - app - Instance of the application under test. If passed, its `testing`
      flag will be set to True and it will be passed on to every TestCase as
      its `app` attribute.
    - db_url - Test database URL.
    - env_env_var - The env var name for specifying the application.
      environment. Only used for the migration error message for now.
    - on_app_ctx_push - A callable invoked upon pushing a test app context.
      Initialize any app context variables in the provided callback.
    - test_dir - The directory containing test suites.
    """

    alembic_config_fn: str = None
    app: Application = None
    db_url: str = None
    env_env_var: str = None
    on_app_ctx_push: Callable = None
    test_dir: str = "./lib/test"

    def run(self):
        args = self._parse_command_line()

        if self.db_url:
            engine = setup_database(self.db_url)

            if self.alembic_config_fn:
                self._check_test_database(engine)

            self._truncate_test_database(engine)

        if self.app:
            self.app.testing = True

        if args.loglevel:
            setup_logging(args.loglevel, args.extra_loglevel)
        else:
            logging.disable()

        if self.on_app_ctx_push:
            receiver = lambda _, **kwargs: self.on_app_ctx_push(**kwargs)
            app_ctx_pushed.connect(receiver)

        self._init_rng(args.random_seed)
        self._run(filter_regex=args.filter, verbose=args.verbose)

    def _run(self, filter_regex=None, verbose=True):
        suite = unittest.defaultTestLoader.discover(self.test_dir, pattern="test_*.py")

        if errors := unittest.defaultTestLoader.errors:
            print(errors[0], file=sys.stderr)

            raise SystemExit(2)

        suite = self._filter_suite(suite, filter_regex)
        runner = unittest.TextTestRunner(verbosity=(2 if verbose else 1))
        exit_value = 0 if runner.run(suite).wasSuccessful() else 1

        raise SystemExit(exit_value)

    def _filter_suite(self, suite, filter_regex):
        filtered_suite = unittest.TestSuite()
        test_cases = self._get_test_cases(suite)
        pattern = re.compile(filter_regex) if filter_regex else None

        for test_case in test_cases:
            nonrunnable = type(test_case).__dict__.get("NONRUNNABLE_BASE_CLASS", False)

            if (not pattern or (pattern.search(test_case.id()) is not None)) and not nonrunnable:
                filtered_suite.addTest(test_case)

        return filtered_suite

    def _get_test_cases(self, suite):

        """ Get TestCases of a TestSuite recursively, in effect flattening the
        structure.
        """

        test_cases = []

        for item in suite:
            if isinstance(item, unittest.TestSuite):
                test_cases += self._get_test_cases(item)
            elif isinstance(item, unittest.TestCase):
                self._set_test_case_ctx(item)
                test_cases.append(item)
            else:
                raise TypeError(
                    "Encountered a bad TestSuite member ({})".format(type(item).__name__))

        random.shuffle(test_cases)

        return test_cases

    def _set_test_case_ctx(self, test_case):

        """ Pass on some context to each test case, only the current app
        instance for now.
        """

        test_case.app = self.app

    def _check_test_database(self, engine):

        """ Bail out if the test database does not exist or is not fully
        migrated.
        """

        alembic_cfg = AlembicConfig(self.alembic_config_fn)
        connection = engine.connect()
        current_head = ScriptDirectory.from_config(alembic_cfg).get_current_head()

        try:
            current_revision = MigrationContext.configure(connection).get_current_revision()
        finally:
            connection.close()

        if current_head != current_revision:
            instructions = ""

            if self.env_env_var:
                instructions = (
                    "\n\nRun:\n\n"
                    f"{self.env_env_var}=\"test\" alembic --config alembic/alembic.ini upgrade head"
                )

            self._err(
                "Test database not accessible or not fully migrated "
                f"(head {current_head} vs revision {current_revision}){instructions}"
            )

    def _truncate_test_database(self, engine):

        """ Truncate all test database tables (except alembic_version). """

        metadata = MetaData()
        connection = engine.connect()
        transaction = connection.begin()

        metadata.reflect(bind=engine)

        for table in metadata.sorted_tables:
            if table.name == "alembic_version":
                continue

            connection.execute(table.delete())

        transaction.commit()
        connection.close()

    def _parse_command_line(self):
        parser = argparse.ArgumentParser(description="Run tests")

        parser.add_argument(
            "-f", "--filter", type=str,
            help="Only run tests whose fully qualified name matches the given regex"
        )
        parser.add_argument(
            "-l", "--loglevel", default=None, type=str,
            help="Enable logging with the given log level"
        )
        parser.add_argument(
            "--extra-loglevel", type=str, help="Log level for extra SQLAlchemy loggers"
        )
        parser.add_argument(
            "-r", "--random-seed", type=int,
            help="Seed for the random number generator"
        )
        parser.add_argument(
            "-v", "--verbose", default=False, action="store_true",
            help="Increase verbosity"
        )

        return parser.parse_args()

    def _init_rng(self, randseed=None):
        if randseed is None:
            randseed = random.randint(0, 9999)

        random.seed(randseed)

        self._print_stderr(f"Using random seed {randseed} for shuffling test cases")

    def _err(self, *args):
        self._print_stderr("ERROR:", *args)
        raise SystemExit(2)

    @staticmethod
    def _print_stderr(*args):
        print(*args, file=sys.stderr)



from contextlib import contextmanager

from flask.ctx import AppContext

from pyvoog.testing.signals import app_ctx_pushed

def setup_app_ctx(app):

    """ Manually set up and push an application context. The caller is
    responsible for tearing down the created context via
    `teardown_app_ctx`.

    This should be the single gateway used for setting up a Flask test app
    context, as we call all receivers listening to the `app_ctx_pushed`
    signal here.
    """

    app_ctx = AppContext(app)
    app_ctx.push()

    app_ctx_pushed.send(None, app_ctx=app_ctx)

    return app_ctx

def teardown_app_ctx(app_ctx):
    app_ctx.pop()

@contextmanager
def app_context(app):

    """ A context manager wrapping `setup_app_ctx` and
    `teardown_app_ctx`.
    """

    try:
        app_ctx = setup_app_ctx(app)
        yield app_ctx
    finally:
        teardown_app_ctx(app_ctx)


import marshmallow

from attrs import define, field
from requests import Response

class AuthenticationError(Exception):
    pass

class NotInitializedError(Exception):
    pass

class ValidationError(marshmallow.ValidationError):
    @property
    def errors(self):
        return self.messages

@define
class ExternalError(Exception):

    """ An error raised due to an external system returning an error
    condition.
    """

    message: str = None
    external_message: str = None
    response: Response = None

class ExternalAuthenticationError(ExternalError):
    pass



from functools import wraps

from pyvoog.controller import Controller, \
    api_endpoint, scoped_endpoint, single_object_endpoint, mutating_endpoint
from pyvoog.db import get_session

class ApiBaseController(Controller):

    """ A base controller class for JSON API resources.

    On your controller class, the following class fields are supported for
    configuration:

    - allowed_actions - a list of allowed actions on this controller.
    - index_order_field - The default field to use for sorting index
      endpoint responses, `id` by default.
    - jwt_secret - base64-encoded JWT secret.
    - model - the backing model of this controller.
    - schema - the Marshmallow schema of acceptable payloads.
    """

    DEFAULT_INDEX_ORDER_FIELD = "id"

    @api_endpoint()
    @scoped_endpoint
    def index(self, query):
        return (
            self.paginate(query, order_by=self._index_order_field, descending=True),
            200,
        )

    @api_endpoint()
    @single_object_endpoint
    def get(self, obj):
        return obj

    @api_endpoint()
    def create(self, *args, **kwargs):
        session, obj = self._create_object(*args, **kwargs)

        session.commit()
        return obj

    @api_endpoint()
    def update(self, *args, **kwargs):
        session, obj = self._update_object(*args, **kwargs)

        session.commit()
        return obj

    @api_endpoint()
    @single_object_endpoint
    def delete(self, obj):
        session = get_session()

        session.delete(obj)
        session.commit()

        return (None, 204)

    @property
    def _index_order_field(self):
        return getattr(self, "index_order_field", self.DEFAULT_INDEX_ORDER_FIELD)

    @mutating_endpoint
    def _create_object(self, payload):
        attrs = self.permit_attributes(self.schema, payload)
        obj = self.model()
        session = get_session()

        for k, v in attrs.items():
            setattr(obj, k, v)

        if getattr(self, "_run_after_model_population", None):
            self._run_after_model_population(obj, payload, action='create')

        session.add(obj)

        return (session, obj)

    @mutating_endpoint
    @single_object_endpoint
    def _update_object(self, obj, payload):
        attrs = self.permit_attributes(self.schema, payload)
        session = get_session()

        for k, v in attrs.items():
            setattr(obj, k, v)

        if getattr(self, "_run_after_model_population", None):
            self._run_after_model_population(obj, payload, action='update')

        session.add(obj)

        return (session, obj)




from pyvoog.util import make_repr

class Endpoint:

    """ Endpoint represents a concrete a path spec and supported HTTP verbs
    mapping to a controller action on a Resource.
    """

    def __init__(self, path, action, methods=["GET"]):
        self.path = path
        self.action = action
        self.methods = methods

    def __repr__(self):
        return make_repr(self)




""" Exports a `GunicornBaseApplication` subclass taking a list of command-
line options for configuration.
"""

from gunicorn.app.base import BaseApplication as GunicornBaseApplication
from gunicorn.config import Config as GunicornConfig

class Server(GunicornBaseApplication):
    def __init__(self, app, argv):
        self.application = app
        self.argv = argv

        super().__init__()

    def load_config(self):
        parser = GunicornConfig(prog="(gunicorn options)").parser()
        args = parser.parse_args(self.argv)

        for k, v in vars(args).items():
            if v is not None and k != "args":
                self.cfg.set(k, v)

    def load(self):
        return self.application



import requests

from attrs import define, field

@define
class UserAgent:

    """ A thin wrapper around Requests automatically adding a set of headers
    to the request. A User-Agent may be added as a shortcut. All other
    arguments to `request` (or rather its wrappers) may be set on the
    instance by `default_rq_args`.
    """

    user_agent: str = None
    headers: dict = {}
    default_rq_args: dict = {}

    def __attrs_post_init__(self):
        if self.user_agent:
            self.headers = self.headers | {"User-Agent": self.user_agent}

    def __getattr__(self, name):
        def make_request(*args, headers={}, **kwargs):
            method = getattr(requests, name)
            headers = self.headers | headers
            extra_rq_kwargs = self.default_rq_args | kwargs

            return method(*args, headers=headers, **extra_rq_kwargs)

        return make_request



import functools
import json
import logging
import re

from datetime import datetime
from urllib.parse import urlparse

import flask as fl
import jwt as pyjwt
import werkzeug.http

from marshmallow import ValidationError as MarshmallowValidationError
from sqlalchemy.exc import NoResultFound
from sqlalchemy import select
from werkzeug.exceptions import BadRequest, MethodNotAllowed

from requests.exceptions import (
    RequestException,
    ConnectionError,
    Timeout,
    TooManyRedirects,
)

from pyvoog.db import get_session

from pyvoog.exceptions import (
    AuthenticationError,
    ExternalError,
    ExternalAuthenticationError,
    ValidationError,
)

from pyvoog.signals import jwt_decoded
from pyvoog.util import AllowException

class _ModelEncoder(json.JSONEncoder):

    """ A JSONEncoder subclass with model instance encoding support. """

    def default(self, obj):
        if hasattr(obj, "as_dict"):
            return obj.as_dict()
        elif isinstance(obj, datetime):
            return self.zulu_isoformat(obj)

        return json.JSONEncoder.default(self, obj)

    @staticmethod
    def zulu_isoformat(d):
        if d.tzinfo:
            return re.sub(r"\+00:00$", "Z", d.isoformat())

        return d.isoformat()

""" ORM-specific controller decorators """

def scoped_endpoint(fn):

    """ A decorator providing the `query` parameter: an SQLAlchemy statement
    with the default scope applied. Also applies `api_endpoint`.
    """

    @functools.wraps(fn)
    def wrapped(self, *args, **kwargs):
        scope = self.model.default_scope()
        query = select(self.model).filter_by(**scope)

        return fn(self, *args, query=query, **kwargs)

    return wrapped

def single_object_endpoint(fn):

    """ A decorator enhancing `scoped_endpoint` and providing the `obj`
    parameter: an object of `self.model`, looked up by the effective default
    scope and incoming ID.
    """

    @functools.wraps(fn)
    def look_up_object(fn):
        def wrapped(self, *args, id, query, **kwargs):
            obj = get_session().execute(query.filter_by(id=id)).scalar_one()
            return fn(self, *args, obj=obj, **kwargs)

        return wrapped

    return functools.update_wrapper(scoped_endpoint(look_up_object(fn)), fn)

""" Generic API facilities """

def api_endpoint(*dec_args, jwt_secret=None, **dec_kwargs):

    """ A high-level API endpoint decorator factory combining
    `json_endpoint`, `emit_http_codes` and `authenticate`. Arguments are
    passed to the `authenticate` decorator factory.

    If `jwt_secret` is not passed in, it is expected to be an attribute on
    the controller.

    In addition, HTTP/405 is raised if the `allowed_actions` itearble
    attribute is present on the controller and does not contain the
    action servicing the request.
    """

    def decorator(fn):
        @functools.wraps(fn)
        def wrapped(self, *args, **kwargs):
            nonlocal dec_kwargs
            nonlocal jwt_secret

            _raise_on_disallowed_action(controller=self, action=fn)

            if jwt_secret is None:
                jwt_secret = self.jwt_secret

            decorated = json_endpoint(
                emit_http_codes(authenticate(*dec_args, jwt_secret=jwt_secret, **dec_kwargs)(fn))
            )

            return decorated(self, *args, **kwargs)

        return wrapped

    return decorator

def json_endpoint(fn):

    """ Decorator for JSONifying API endpoints. By default the result is
    encoded and returned with the 200 status code. A Response object is
    passed through. If a tuple is returned from the wrapped routine:

    - if its length is one, the element indicates a HTTP status code and the
      payload becomes the matching status string
    - if its length is >1, the first element is the payload, the second is
      the HTTP status code and the optional third element is a dict of extra
      headers.
      """

    @functools.wraps(fn)
    def wrapped(self, *args, **kwargs):
        res = fn(self, *args, **kwargs)
        res_is_tuple = type(res) is tuple
        headers = {"Content-Type": "application/json"}
        payload = res
        code = 200

        if type(res) is fl.Response:
            return res
        elif res_is_tuple and len(res) == 1:
            code = res[0]
            payload = werkzeug.http.HTTP_STATUS_CODES[code]
        elif res_is_tuple:
            payload = res[0]
            code = res[1]

            with AllowException(IndexError):
                if res[2] is not None:
                    headers |= res[2]

        return (json.dumps(payload, cls=_ModelEncoder), code, headers)

    return wrapped

def emit_http_codes(fn):

    """ Turn errors into HTTP/4xx responses:

    - BadRequest — HTTP/400
    - AuthenticationError — HTTP/401
    - None return value or a NoResultFound exception — HTTP/404
    - ValidationError (pyvoog or vanilla Marshmallow) — HTTP/422 with a
      payload describing the errors in `errors`.
    - NotImplementedError — HTTP/501.

    See also `handle_upstream_errors` for a complementary decorator.
    """

    @functools.wraps(fn)
    def wrapped(self, *args, **kwargs):
        try:
            res = fn(self, *args, **kwargs)
        except AuthenticationError as e:
            res = get_response_tuple(401, str(e))
        except NoResultFound:
            res = get_response_tuple(404)
        except BadRequest:
            res = get_response_tuple(400)
        except ValidationError as e:
            res = (dict(errors=e.errors), 422)
        except MarshmallowValidationError as e:
            res = (dict(errors=e.normalized_messages()), 422)
        except NotImplementedError:
            res = get_response_tuple(501)

        return get_response_tuple(404) if res is None else res

    return wrapped

def authenticate(jwt_secret):

    """ The returned decorator raises AuthenticationError on authentication
    failure and emits the `jwt_decoded` signal with the decoded JWT payload
    on success. The `exp` claim is currently required unconditionally on the
    token and stale tokens are rejected.
    """

    def decorator(fn):
        @functools.wraps(fn)
        def wrapped(self, *args, **kwargs):
            jwt = _get_jwt_from_request()

            try:
                payload = pyjwt.decode(
                    jwt, jwt_secret, algorithms="HS256", options=dict(require=["exp"])
                )
            except Exception as e:
                logging.warn(f"Authentication failure for token \"{jwt}\": {e}")
                raise AuthenticationError("Not Authenticated")

            jwt_decoded.send(fl.current_app, payload=payload)
            return fn(self, *args, **kwargs)

        return wrapped

    return decorator

def mutating_endpoint(fn):

    """ A decorator providing the `payload` parameter containing the
    deserialized incoming JSON.
    """

    @functools.wraps(fn)
    def wrapped(self, *args, **kwargs):
        return fn(self, *args, payload=fl.request.get_json(), **kwargs)

    return wrapped

def handle_upstream_errors(fn):

    """ Decorator to turn ExternalAuthenticationErrors into HTTP/401
    responses, all other ExternalErrors into HTTP/502 responses and selected
    Requests exceptions into detailed HTTP/5xx responses.

    See also `emit_http_codes` for a complementary decorator.
    """

    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except ExternalError as e:
            extra_args = dict(external_message=e.external_message) if e.external_message else {}
            code = 401 if isinstance(e, ExternalAuthenticationError) else 502

            return get_response_tuple(code, e.message, **extra_args)
        except RequestException as e:
            url = e.request.url
            host = urlparse(url).hostname

            if isinstance(e, Timeout):
                code = 504
                message = f"Timed out connecting to {host}"
            elif isinstance(e, ConnectionError):
                code = 500
                message = f"Failed connecting to {host}"
            elif isinstance(e, TooManyRedirects):
                code = 502
                message = f"The request to {url} resulted in too many redirects"
            else:
                raise e

            return get_response_tuple(code, message, details=str(e))

    return wrapped

def get_response_tuple(code, /, message=None, **kwargs):

    """ Utility routine to generate a standard error response payload and
    return it as a pair along with the HTTP error code.
    """

    payload = {
        "message": message or werkzeug.http.HTTP_STATUS_CODES[code],
        **kwargs
    }

    return (payload, code)

def _get_jwt_from_request():
    if not (jwt := fl.request.args.get("token")):
        try:
            jwt = re.split(r"Bearer\s+", fl.request.headers.get("Authorization", ""))[1]
        except Exception:
            jwt = None

    return jwt

def _raise_on_disallowed_action(controller, action):
    allowed_actions = getattr(controller, "allowed_actions", None)

    if (allowed_actions is not None) and (action.__name__ not in controller.allowed_actions):
        raise MethodNotAllowed()



import collections
from collections.abc import Mapping, MutableMapping

class VanillaDict(dict):

    """ A helper class for data structures passed into
    mapping_to_namedtuple. VanillaDicts are not converted into named tuples
    and remain dicts.
    """

def mapping_to_namedtuple(mapping, class_name="MappingToNamedtuple"):

    """ Recursively convert a mapping to a namedtuple. """

    nt = collections.namedtuple(class_name, mapping.keys())(**mapping)

    for (k, v) in mapping.items():
        is_mapping = isinstance(v, Mapping) or isinstance(v, MutableMapping)

        if not is_mapping or isinstance(v, VanillaDict):
            continue
        nt = nt._replace(**{k: mapping_to_namedtuple(v, class_name)})

    return nt

def filter_mapping(fn, mapping):

    """ A `filter` counterpart for mapping types, with the same signature
    and parameter semantics (as far as applicable to mappings). Filtering
    applies to values.

    See:

    https://docs.python.org/3/library/functions.html#filter
    """

    if fn is None:
        fn = lambda x: x

    return {k: v for k, v in mapping.items() if fn(v)}




""" Custom logging utilities. """

import logging

from functools import partial

import flask as fl

class PrefixedLogRecord(logging.LogRecord):

    """ A LogRecord subclass providing the `prefix` field containing the
    logger name, unless it's the root logger.
    """

    def __init__(self, name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.prefix = "" if name == "root" else "[{}] ".format(name)

class MultilineLogRecord(PrefixedLogRecord):

    """ A LogRecord subclass splitting incoming messages on newlines and
    converting each to a LogRecord of type `constituent_cls`
    """

    def __init__(self, name, level, fn, lno, msg, *args, constituent_cls=None, **kwargs):
        self.constituents = (
            constituent_cls(name, level, fn, lno, line, *args, *kwargs) for line in msg.split("\n")
        )

        super().__init__(name, level, fn, lno, msg, *args, **kwargs)

class MultilineFormatter(logging.Formatter):

    """ A formatter for MultilineLogRecords, formatting each constituent
    record independently and returning the concatenated result.
    """

    def format(self, record):
        formatted_lines = []

        if not isinstance(record, MultilineLogRecord):
            raise TypeError(
                "MultilineFormatter can only format MultilineLogRecords, "
                f"but received {type(record).__name__}"
            )

        for line_record in record.constituents:
            super_instance = super(MultilineFormatter, self)

            try:
                formatted_lines.append(super_instance.format(line_record))
            except TypeError:

                # When breaking on newlines, message placeholders may not be present in
                # the resulting constituent records, resulting in formatting failures.
                # Fall back to formatting the entire record in such cases.

                formatted_lines = (super_instance.format(record),)
                break

        return "\n".join(formatted_lines)

class ContextfulLogger:

    """ A convenience class providing logging methods matching the level
    names in `logging`. Invocations are forwarded to `logging.log` with the
    appropriate level. A prefix constructed from the positional and keyword
    args to the constructor (modifiable by `amend_context`) is prepended to
    each log message.
    """

    LOGGABLE_LEVELS = [
        "critical",
        "error",
        "warning",
        "info",
        "debug"
    ]

    def __init__(self, *args, **kwargs):
        self.ctx_args = list(args)
        self.ctx_kwargs = kwargs

        self._prefix = self._compile_prefix()

    def amend_context(self, *args, **kwargs):
        self.ctx_args.extend(args)
        self.ctx_kwargs |= kwargs

        self._prefix = self._compile_prefix()

    def _log(self, level, msg, *args, **kwargs):
        logging.log(level, f"{self._prefix} {msg}", *args, **kwargs)

    def _compile_prefix(self):
        args_prefix = ', '.join(self.ctx_args)
        kwargs_prefix = ', '.join(f"{k}={v}" for k, v in self.ctx_kwargs.items())
        separator = ", " if args_prefix and kwargs_prefix else ""

        return f"[{args_prefix}{separator}{kwargs_prefix}]"

    def __getattr__(self, name):
        if name in self.LOGGABLE_LEVELS:
            return partial(self._log, getattr(logging, name.upper()))

        raise AttributeError(f"Cannot forward `{name}` to the `logging` module")

def setup_logging(level_str, extra_level_str, custom_extra_loggers=()):

    """ Set up logging with timestamped and prefixed log records, allowing
    log level differentiation for SQLAlchemy loggers.
    """

    root_logger = logging.getLogger()
    level = getattr(logging, level_str.upper())
    formatter = MultilineFormatter("%(asctime)s %(levelname)7s: %(prefix)s%(message)s")

    extra_loggers = (
        "sqlalchemy.pool",
        "sqlalchemy.dialects",
        "sqlalchemy.orm",
        "sqlalchemy.engine",
    )

    logging.setLogRecordFactory(make_log_record)

    if not root_logger.handlers:
        logging.basicConfig()

    root_logger.setLevel(level)
    root_logger.handlers[0].setFormatter(formatter)

    if extra_level_str:
        extra_level = getattr(logging, extra_level_str.upper())

        for logger in extra_loggers + custom_extra_loggers:
            logging.getLogger(logger).setLevel(extra_level)

def make_log_record(name, level, fn, lno, msg, args, exc_info, func=None, sinfo=None, **kwargs):

    """ LogRecord factory. Note that path is currently passed as None.
    Implement path name extraction when desired.
    """

    return MultilineLogRecord(
        name, level, None, lno, msg, args, exc_info, func, sinfo, constituent_cls=PrefixedLogRecord
    )

def log_requests(app, make_log_string=None):

    """ Call with an application instance to register an `after_request`
    handler logging all requests.
    """

    if make_log_string is None:
        make_log_string = make_request_log_string

    def log_request(response):
        request = fl.request

        if make_log_string:
            message = make_log_string(request, response)
        else:
            message = (
                f"Completed {request.method} {request.path} for {request.remote_addr} "
                f"with {response.status}"
            )

        logging.info(message)

        return response

    app.after_request(log_request)

def make_request_log_string(request, response):
    return (
        f"Completed {request.method} {request.path} for {request.remote_addr} "
        f"with {response.status}"
    )

def get_logger_level(name=None):
    return logging.getLogger(name).level



import importlib
import logging
import re

import flask as fl

from attrs import define
from stringcase import snakecase

from .endpoint import Endpoint
from .namespace import Namespace
from .resource import Resource
from .util import normalize_path

@define
class Router:
    controller_ns: str

    DEFAULT_ENDPOINTS_TEMPLATE = [
        dict(path="{}s", methods=["GET"], action="index"),
        dict(path="{}s", methods=["POST"], action="create"),
        dict(path="{}/<int:id>", methods=["GET"], action="get"),
        dict(path="{}/<int:id>", methods=["PUT"], action="update"),
        dict(path="{}/<int:id>", methods=["DELETE"], action="delete")
    ]

    def route(self, table):

        """ Route requests to controllers based on the incoming iterable routing
        table containing Resources (optionally wrapped in Namespaces), in turn
        containing Endpoints specifying the actions handling matching requests.
        A Resource is not required to contain explicit endpoints — if no
        endpoints are configured, a default set of RESTful endpoints for the
        resource are set up.

        The controller module to import from the namespace specified by
        `controller_ns` is inferred from the path prefix and resource name. The
        module is expected to contain a class with a name constructed by
        titlecasing the resource name and appending "Controller" to it, which is
        instantiated with the app object. The controller must contain a method
        for every action specified in Endpoints for serving the specified
        request.
        """

        for namespace_or_resource in table:
            if isinstance(namespace_or_resource, Namespace):
                resources = namespace_or_resource.resources
                path_prefix = namespace_or_resource.path_prefix
            elif isinstance(namespace_or_resource, Resource):
                resources = (namespace_or_resource,)
                path_prefix = "/"
            else:
                raise TypeError(
                    "Expected a Namespace or Resource in the routing table, "
                    f"but received {namespace_or_resource}"
                )

            for resource in resources:
                self._route_resource(path_prefix, resource)

    def _route_resource(self, path_prefix, resource):
        if not isinstance(resource, Resource):
            raise TypeError(
                f"Expected a Resource in router config, but received {resource}"
            )

        module = self._import_controller(path_prefix=path_prefix, resource=resource)
        controller_cls = getattr(module, resource.ctrlr_class_name)
        controller = controller_cls()
        endpoints = resource.endpoints if resource.endpoints else []

        if not endpoints or resource.include_default_endpoints:
            endpoints += self._populate_default_endpoints(resource.name)

        for endpoint in endpoints:
            if not isinstance(endpoint, Endpoint):
                raise TypeError(
                    f"Expected an Endpoint in Resource config, but received {endpoint}"
                )

            self._route_to_controller(controller, path_prefix, endpoint)

    def _route_to_controller(self, controller, path_prefix, endpoint):

        """ Route paths (path prefix + endpoint path) to controller actions. """

        path = normalize_path(f"{path_prefix}/{endpoint.path}")
        func = getattr(controller, endpoint.action)
        ctrlr_name = type(controller).__name__
        endpoint_name = snakecase(f"{ctrlr_name}_{endpoint.action}")

        logging.info(
            f"Adding route: {path} -> {ctrlr_name}.{endpoint.action} "
            f"({','.join(endpoint.methods)})"
        )

        fl.current_app.add_url_rule(
          path,
          view_func=func,
          endpoint=endpoint_name,
          methods=endpoint.methods
        )

    def _import_controller(self, path_prefix, resource):
        stripped_path_prefix = re.sub(r"^/+", "", path_prefix)
        ns_infix = f".{stripped_path_prefix.replace('/', '.')}." if stripped_path_prefix else "."
        module_name = f"{self.controller_ns}{ns_infix}{resource.ctrlr_module_name}"

        return importlib.import_module(module_name)

    def _populate_default_endpoints(self, name):
        return map(
            lambda kws: Endpoint(**(kws | {"path": kws["path"].format(name)})),
            self.DEFAULT_ENDPOINTS_TEMPLATE
        )



import os
import os.path
import sys

from functools import reduce

def drop_last_path_components(path, n):

    """ Given a path, normalize it using `abspath`, drop its last `n`
    components and return the result.
    """

    return reduce(lambda acc, _: os.path.split(acc)[0], range(n), os.path.abspath(path))

def set_root_directory(dirname):

    """ Change cwd to `dirname` and append it to `sys.path`. Useful for
    initializing scripts to work in the project's environment. Passes
    through `dirname` for convenience.
    """

    os.chdir(dirname)

    if dirname not in sys.path:
        sys.path.append(dirname)

    return dirname



import importlib
import inspect

class AllowException:

    """ Prefer `contextlib.suppress` as a builtin alternative. """

    def __init__(self, *excs):
        self.excs = excs

    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_val, exc_tb):
        if (exc_type and any(issubclass(exc_type, our_exc_type) for our_exc_type in self.excs)):
            return True
        return False

class Undefined:

    """ A class to convey a null value, distinct from None. """

def make_namespace_importer(module_template, subclass_of, return_class=False):

    """ A function returning factory receiving a code, importing a module
    based on the code, extracting a class from the imported module and
    returning an instance of the class.

    - module_template - A module name format string containing the `code`
      placeholder, filled in by the factory for the resolved module name.
    - subclass_of - A class acting as a filter - the imported module's
      members are scanned and the first found class that is a strict
      subclass of `subclass_of` is used for instantiation.
    - return_class - Instead of instantiating the class, return the
      class itself.

    importlib raises a ModuleNotFoundError if the expected module cannot be
    loaded; an ImportError is raised if the specified subclass is not
    present in the module.
    """

    def cls_filter(m):
        return inspect.isclass(m) \
            and issubclass(m, subclass_of) \
            and m is not subclass_of

    def import_by_code(code, *args, **kwargs):
        module_name = module_template.format(code=code)
        module = importlib.import_module(module_name)

        try:
            cls = inspect.getmembers(module, cls_filter)[0][1]
        except IndexError:
            raise ImportError(f"Cannot find a {subclass_of.__name__} subclass in {module_name}")

        if return_class:
            return cls

        return cls(*args, **kwargs)

    return import_by_code

def make_repr(obj):

    """ Generate an informative repr for `obj`. Consider using attrs, as it
    provides a nice repr as a bonus.
    """

    attrs = ", ".join(f"{k}={v}" for k, v in vars(obj).items())

    return f"<{type(obj).__name__}({attrs})>"





""" A Flask subclass setting up request logging and responding to HTTP
errors with a JSON payload.
"""

import flask as fl
import werkzeug.http

from pyvoog.db import teardown_sessions
from pyvoog.controller import get_response_tuple
from pyvoog.logging import log_requests
from pyvoog.util import AllowException

class Application(fl.Flask):
    def __init__(self, name=None):
        super().__init__(name or self.__class__.__name__)

        with self.app_context():
            app = fl.current_app

            log_requests(app)
            app._register_error_handlers()
            app._register_teardown_funcs()

            if getattr(self, "__app_post_init__", None):
                self.__app_post_init__()

    def _register_error_handlers(self):

        """ Register error handlers for all valid 4xx and 5xx HTTP status codes
        to return JSON.
        """

        def get_handler(code):
            return lambda _: fl.make_response(
                *get_response_tuple(code),
                {"Content-Type": "application/json"}
            )

        for code, _ in werkzeug.http.HTTP_STATUS_CODES.items():
            if code >= 400:
                with AllowException(KeyError, ValueError):
                    self.register_error_handler(code, get_handler(code))

    def _register_teardown_funcs(self):
        self.teardown_appcontext(teardown_sessions)



import re

from stringcase import snakecase

import flask as fl
import marshmallow.exceptions

from sqlalchemy import select, and_, or_

from pyvoog.db import get_session
from pyvoog.exceptions import ValidationError

class Controller:
    DEFAULT_PER_PAGE = 50
    MAX_PER_PAGE = 250

    def permit_attributes(self, schema, payload):

        """ Validate an incoming payload against a Marshmallow schema, return
        the loaded result. Cast validation errors to our ValidationError
        subclass (also thrown on model validation).
        """

        try:
            return schema.load(payload)
        except marshmallow.exceptions.ValidationError as e:
            e.__class__ = ValidationError
            raise

    def paginate(self, query, order_by, descending=False, payload_key=None):

        """ Given an SQLAlchemy statement (`query`) and the name of the column
        determining ordering, paginate output and return a dict. The key of the
        objects array is passed in `payload_key` or deduced automatically from
        the controller name. Pagination metadata is included in `pagination`. If
        `from` is passed in the HTTP query string, this is the ID of the object
        to start output from. The `per_page` parameter controls the maximum
        number of items to output, up to MAX_PER_PAGE. If any further items
        remain after the those output, the ID of the next object is returned in
        the `pagination.next_cursor`. This can be used as the value of `from` to
        the next request.
        """

        model = self.model
        per_page = self._items_per_page
        from_id = fl.request.args.get("from", None)
        ordering_column = getattr(model, order_by)
        empty_result = False
        cursor = None

        query = (
            query
            .order_by(ordering_column.desc() if descending else ordering_column, self.model.id)
            .limit(per_page + 1)
        )

        if from_id:
            query = self._start_pagination_at(query, from_id, ordering_column, descending)

        try:
            *body, _next = get_session().execute(query).scalars()
        except ValueError:
            empty_result = True

        if empty_result:
            payload = []
        elif len(body) < per_page:
            payload = [*body, _next]
        else:
            payload = body
            cursor = _next.id

        if not payload_key:
            payload_key = f'{snakecase(re.sub(r"Controller$", "", self.__class__.__name__))}s'

        return {
            payload_key: payload,
            "pagination": {
                "next_cursor": cursor
            }
        }

    def _start_pagination_at(self, query, from_id, ordering_column, descending):
        model = self.model
        milestone_value = select(ordering_column).where(model.id == from_id).scalar_subquery()
        criterion = ordering_column > milestone_value

        if descending:
            criterion = ordering_column < milestone_value

        return query.where(
            or_(
                criterion,
                and_(ordering_column == milestone_value, model.id >= from_id),
            )
        )

    @property
    def _items_per_page(self):
        try:
            per_page = int(fl.request.args.get("per_page") or self.DEFAULT_PER_PAGE)
        except Exception:
            per_page = self.DEFAULT_PER_PAGE

        return min(max(1, per_page), self.MAX_PER_PAGE)




""" Sundry command-line argument parsing routines. Provides a set of
common recognized arguments and allows passing through arguments to
Gunicorn verbatim, separated from application args by `--`.
"""

import argparse
import sys

from attrs import define, field

@define
class Args:
    parser_args: dict = {}
    _parser: argparse.ArgumentParser = field(init=False, default=None)
    _parsed_args: argparse.Namespace = field(init=False, default=None)
    _our_argv: list = field(init=False)
    _rest_of_argv: list = field(init=False)

    def __attrs_post_init__(self):
        (self._our_argv, self._rest_of_argv) = self._split_command_line()

    @property
    def parser(self):
        if not self._parser:
            defaults = dict(
                epilog="Any arguments specified after `--` are passed on to the gunicorn server."
            )

            self._parser = argparse.ArgumentParser(**(defaults | self.parser_args))

        return self._parser

    @property
    def parsed_args(self):
        if not self._parsed_args:
            self._parsed_args = self.parser.parse_args(self._our_argv)

        return self._parsed_args

    @property
    def gunicorn_argv(self):

        """ Return the portion of argv separated by `--`. Translate our
        arguments to gunicorn equivalents and append to the former. Note that
        this will parse arguments, if this has not happened alredy.
        """

        args = self.parsed_args
        gunicorn_argv = self._rest_of_argv.copy()

        if args.port:
            gunicorn_argv += ["--bind", f"{args.bind}:{args.port}"]

        return gunicorn_argv

    def add_common_argumets(self, **defaults):

        """ Add default arguments to the parser. Override in a subclass to add
        application-specific arguments.
        """

        parser = self.parser

        parser.add_argument(
            "-b", "--bind", default="127.0.0.1", type=str,
            help="The interface to bind to, 127.0.0.1 by default"
        )
        parser.add_argument(
            "-p", "--port", default=defaults.get("port"), type=int,
            help="The port to listen on, {} by default".format(defaults.get("port"))
        )
        parser.add_argument(
            "-d", "--database", default=defaults.get("database"), type=str,
            help="The database URL, {} by default".format(defaults.get("database"))
        )
        parser.add_argument(
            "-l", "--loglevel", default=defaults.get("loglevel"), type=str,
            help="Log level, {} by default".format(defaults.get("loglevel"))
        )
        parser.add_argument(
            "--extra-loglevel", type=str, help="Log level for extra SQLAlchemy loggers"
        )
        parser.add_argument(
            "--hide-sql-params", action="store_true", help="Hide SQL parameters from log entries"
        )

    @staticmethod
    def _split_command_line():

        """ Split the command line on `--` and return the lists of
        arguments.
        """

        args = sys.argv[1:]
        separator_idx = None

        for i, arg in enumerate(args):
            if arg == "--":
                separator_idx = i
                break

        if separator_idx is None:
            return (args, [])

        return (args[:separator_idx], args[separator_idx + 1:])



from collections.abc import Mapping
from contextlib import contextmanager
from unittest import TestCase

from pyvoog.db import get_plain_session
from pyvoog.testing.util.models import temporary_object
from pyvoog.testing.util.requests import controller_fixture

class ControllerTestCase(TestCase):

    """ A controller test suite base class, providing convenience
    routines for making typical requests. Respects the following
    attributes on the object:

    - ENDPOINT - the default endpoint for making requests.
    - jwt_secret - JWT secret for constructing the access token.
    - jwt_payload - JWT payload for constructing the access token.
    """

    @contextmanager
    def get_response(self, model, model_args, endpoint=None):

        """ A context manager constructing a temporary instance of `model`
        with `model_args`, making a GET request to an endpoint and yielding
        the response.

        `endpoint` may be a string or a callable (in which case the callable is
        invoked with `obj` and its result used as the effective endpoint).
        `self.ENDPOINT` is used as the default callback.
        """

        with self.bound_controller_fixture as ua:
            with temporary_object(model, **model_args) as obj:
                if callable(endpoint):
                    effective_endpoint = endpoint(obj)
                else:
                    effective_endpoint = endpoint or self.ENDPOINT

                yield ua.get(effective_endpoint)

    @contextmanager
    def post_response(self, model=None, payload=None):

        """ A context manager making a POST request to `self.ENDPOINT` and
        yielding the response.

        If the request is successful, `model` is present and the response body
        is a JSON object containing `id`, the persisted object is cleaned up.
        """

        with self.bound_controller_fixture as ua:
            response = ua.post(self.ENDPOINT, json=payload)
            session = get_plain_session()

            try:
                yield response
            finally:
                json = response.json
                id = isinstance(json, Mapping) and json.get("id", None)

                if model and (id is not None) and response.status_code < 400:
                    session.delete(session.get(model, id))
                    session.commit()

    @property
    def bound_controller_fixture(self):

        """ Return a `controller_fixture` context manager bound to the app under
        test and authentication credentials.
        """

        return controller_fixture(
            self.app, jwt_secret=self.jwt_secret, jwt_payload=self.jwt_payload
        )



import logging

from collections import namedtuple
from contextlib import contextmanager
from itertools import chain

import flask as fl

from sqlalchemy import create_engine, engine, event
from sqlalchemy.orm import Session

from pyvoog.exceptions import NotInitializedError

_PerRequestSession = namedtuple('_PerRequestSession', ['value'])

_engine = None

class ValidatingSession(Session):

    """ A Session automatically attaching a before_flush hook to run
    validations on all models.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        event.listen(self, "before_flush", self.__class__.run_validations)

    @staticmethod
    def run_validations(session, flush_context, instances):
        for obj in chain(session.new, session.dirty):
            obj.validate()

def setup_database(db_url, **kwargs):
    global _engine

    _engine = create_engine(db_url, echo=False, future=True, pool_pre_ping=True, **kwargs)

    return _engine

def get_session(key="session", cls=ValidatingSession):

    """ Return a per-request SQLAlchemy Session, creating one if needed.
    There may be several active sessions, differentiated by the given key.
    The teardown listener must be registered separately at app
    initialization, as this is no longer allowed once a request is in
    progress.
    """

    if key not in fl.g:
        logging.debug(f"Setting up per-request session '{key}'")

        if not isinstance(_engine, engine.Engine):
            raise NotInitializedError("Database engine has not been set up.")

        setattr(fl.g, key, _PerRequestSession(cls(_engine)))

    return fl.g.get(key).value

def get_plain_session():

    """ As `get_session`, but yield a vanilla Session instance. """

    return get_session(key="plain_session", cls=Session)

def teardown_sessions(exc):
    session_keys = list(filter(lambda k: isinstance(fl.g.get(k), _PerRequestSession), fl.g))

    for key in session_keys:
        logging.debug(f"Tearing down per-request session '{key}'")

        session = fl.g.pop(key).value
        session.close()

@contextmanager
def temporary_session(cls=ValidatingSession):
    session = cls(_engine)

    try:
        yield session
    finally:
        session.close()




from attrs import define, field

from .util import normalize_path

@define
class Namespace:

    """ Namespace encapsulates one or more Resources with a common path
    prefix. Path prefixes are normalized.
    """

    path_prefix = field(converter=normalize_path)
    resources = field(factory=list)

    def __init__(self, path_prefix, *args):
        self.__attrs_init__(path_prefix=path_prefix, resources=args)



from marshmallow import Schema, EXCLUDE, RAISE

def make_schema(**d):

    """ Create a Marshmallow schema instance from the incoming dict. Permit
    unknown keys and exclude these.
    """

    return Schema.from_dict(d)(unknown=EXCLUDE)

def make_strict_schema(**d):

    """ As `make_schema`, but do not allow unknown keys. """

    return Schema.from_dict(d)(unknown=RAISE)



from pyvoog.exceptions import NotInitializedError
from pyvoog.util.mapping import mapping_to_namedtuple

import importlib
import os
import re

class ConfigLoader:

    """ A class for loading the configuration from config.*, according
    to the environment passed in the environment variable `envvar`. The
    default environment is always loaded; the secondary environment
    overwrites the default configuration values.

    If `overwrite_prefix` is specified, all configuration values can be
    overridden using environment variables. The name of the environment
    variable must be the overwrite prefix and configuration key combined,
    uppercased. For instance, if the overwrite prefix is "MY_APP", the
    environment variable "MY_APP_CACHE_BACKEND" will take precedence over
    the `cache_backend` configuration value.
    """

    CONFIG_PKG = "config"

    def __init__(self, envvar=None, env=None, overwrite_prefix=None):
        config = self._load_env("default")
        env_from_env = envvar and os.environ.get(envvar, None)
        effective_env = env_from_env or env

        if effective_env:
            config.update(self._load_env(effective_env))

        self.env = env
        self.config = config
        self.overwrite_prefix = overwrite_prefix

    def load(self):
        global _config

        _config = mapping_to_namedtuple(self._overwrite_from_os_env(self.config), "Config")

        return config

    def _load_env(self, env):
        return importlib.import_module("{}.{}".format(self.CONFIG_PKG, env)).config

    def _overwrite_from_os_env(self, config):
        prefix = self.overwrite_prefix

        if prefix is not None:
            var_name_regex = fr"{prefix}(\w+)$"
            overwrites = {}

            for k, v in os.environ.items():
                if m := re.match(var_name_regex, k):
                    config_key = m[1].lower()
                    overwrites[config_key] = v

            config.update(overwrites)

        return config

class _Config:
    @staticmethod
    def __contains__(key):
        return hasattr(_config, key)

    @staticmethod
    def __getattr__(name):
        if not _config:
            raise NotInitializedError("Configuration has not been initialized")

        return getattr(_config, name)

    @staticmethod
    def __getitem__(key):
        if hasattr(_config, key):
            return getattr(_config, key)

        raise KeyError(key)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

_config = None
config = _Config()



from contextlib import contextmanager

from pyvoog.db import get_session

@contextmanager
def temporary_object(model, persist=False, **kwargs):

    """ Temporarily persist a model instance, making it available within a
    context manager and discarding it after use.
    """

    session = get_session()
    obj = create_object(model, session=session, **kwargs)

    try:
        yield obj
    finally:
        if not persist:
            session.delete(obj)
            session.commit()

def create_object(model, session=None, **kwargs):
    session = session or get_session()
    obj = initialize_object(model, **kwargs)

    session.add(obj)
    session.commit()

    return obj

def initialize_object(model, **kwargs):
    obj = model()

    for k, v in kwargs.items():
        setattr(obj, k, v)

    return obj

def delete_object(obj, session=None):
    session = session or get_session()

    session.delete(obj)
    session.commit()


from contextlib import contextmanager
from time import time

import flask as fl
import jwt

from flask.testing import FlaskClient
from werkzeug.test import EnvironBuilder

from .util import app_context

class _AuthenticatedClient(FlaskClient):
    def __init__(
        self, *args, headers={}, jwt_payload=None, jwt_secret=None, set_jwt_exp=True, **kwargs
    ):
        self.headers = headers

        if jwt_payload and jwt_secret:
            if set_jwt_exp:
                jwt_payload = jwt_payload | dict(exp=time() * 1.1)

            token = jwt.encode(jwt_payload, jwt_secret)
            self.headers = self.headers | {"Authorization": f"Bearer {token}"}

        super().__init__(*args, **kwargs)

    def open(self, *args, **kwargs):
        return super().open(*args, headers=self.headers, **kwargs)

@contextmanager
def controller_fixture(app, **client_kwargs):

    """ Yield a context manager setting up a Flask app context and a test
    client for use in controller testing. The context variable is a
    FlaskClient subclass instance intended to make convenient authenticated
    requests. Pass `jwt_secret` and `jwt_payload` to construct a token,
    included with every request. If `set_jwt_exp` is true, a future
    timestamp is appended to the payload. Extra custom HTTP headers may be
    included via `headers`.
    """

    app.test_client_class = _AuthenticatedClient

    with app_context(app):
        yield app.test_client(**client_kwargs)

def build_request(*args, **kwargs):

    """ Forward arguments to EnvironBuilder and return a Flask Request based
    on the constructed env.
    """

    builder = EnvironBuilder(*args, **kwargs)

    return fl.Request(builder.get_environ())



from functools import wraps

from sqlalchemy import select, Column

from pyvoog.db import get_plain_session
from pyvoog.exceptions import ValidationError

def requires_column(ctor):

    """ Decorator to enforce validator application to SQLAlchemy Columns
    exclusively.
    """

    @wraps(ctor)
    def wrapped(self, column, *args, **kwargs):
        if not isinstance(column, Column):
            raise TypeError(f"The {type(self).__name__} validator may only be applied to a Column")

        return ctor(self, column, *args, **kwargs)

    return wrapped

class Uniqueness:

    """ Enforce a uniqueness constraint on a column. Test if a database
    record with the given value exists, fail if this is the case. A scope
    may optionally be passed as a list of column names.

    This validator requires an `sqlalchemy.Column` as the validatable.
    """

    @requires_column
    def __init__(self, column, scope=[]):
        self.column = column
        self.scope = scope

    def run(self, obj):
        value = getattr(obj, self.column.name)
        id_column = type(obj).__table__.c['id']
        scope_dict = {col_name: getattr(obj, col_name) for col_name in self.scope}
        query = select(True).select_from(obj.__class__) \
            .filter_by(**scope_dict).where(self.column == value, id_column != obj.id)
        row = get_plain_session().execute(query).first()

        if row is not None:
            raise ValidationError(["Is not unique."])

class Inclusion:

    """ Check that a value is included in the given list, fail
    otherwise.
    """

    def __init__(self, attr, belongs_to=[]):
        self.attr = attr
        self.belongs_to = belongs_to

    def run(self, obj):
        value = getattr(obj, self.attr.name)

        if value not in self.belongs_to:
            raise ValidationError(["Does not belong to the set of allowed values."])

class Schema:

    """ Validate a value against a Marshmallow schema. """

    def __init__(self, attr, schema, required=False):
        self.attr = attr
        self.schema = schema
        self.required = required

    def run(self, obj):
        value = getattr(obj, self.attr.name)

        if not value and self.required:
            raise ValidationError("Must be present.")
        elif errors := self.schema().validate(value):
            raise ValidationError(errors)

class Custom:

    """ Delegate validation to a method on the model. """

    def __init__(self, attr, validator_name):
        self.validator_name = validator_name

    def run(self, obj):
        getattr(obj, self.validator_name)()



import re

def normalize_path(path_prefix: str) -> str:

    """ Normalize a path, collapsing any slashes within the path and
    removing any slashes at the beginning and end of the path.
    """

    return re.sub(r"^/*|/+", "/", re.sub(r"/+$", "", path_prefix or ""))



from attrs import define, field
from stringcase import pascalcase

@define
class Resource:

    """ A Resource represents a collection of related objects accessible via
    the API, essentially mapping to a controller.

    - name - Used for determining the controller module, the controller
      class name and the path prefix for default endpoints.
    - ctrlr_class_name - Explicitly set the name of controller class mapping
      to the Resource.
    - ctrlr_class_suffix - Set the controller class name suffix when
      inferring the class name from `name`, "Controller" by default.
    - ctrlr_module_name - Explicitly set the suffix of the controller module
      mapping to the Resource. The router's `controller_ns` and any
      encapsulating namespace's path prefix are nevertheless respected.
    - ctrlr_module_suffix - Set the module name suffix when inferring the
      module name from `name`, "_controller" by default.
    """

    name: str = None
    endpoints: list = []
    include_default_endpoints: bool = False
    ctrlr_class_name: str = None
    ctrlr_class_suffix: str = "Controller"
    ctrlr_module_name: str = None
    ctrlr_module_suffix: str = "_controller"

    def __attrs_post_init__(self):
        if not self.name and (
            not self.ctrlr_class_name
            or not self.ctrlr_module_name
            or self.include_default_endpoints
        ):
            raise ValueError(
                "`ctrlr_module_name` and `ctrlr_class_name` are required, and "
                "`include_default_endpoints` must be False if `name` is not provided for a Resource"
            )

        if not self.ctrlr_class_name:
            self.ctrlr_class_name = f"{pascalcase(self.name)}{self.ctrlr_class_suffix}"

        if not self.ctrlr_module_name:
            self.ctrlr_module_name = f"{self.name}{self.ctrlr_module_suffix}"



import itertools
import types

import sqlalchemy

from datetime import datetime, timezone
from deepmerge import always_merger
from marshmallow import Schema, fields, validate
from sqlalchemy.event import listen
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy import Column, inspect, select, types as sa_types
from sqlalchemy.orm import declared_attr, declarative_base, object_session
from sqlalchemy.sql import func
from sqlalchemy.sql.sqltypes import Boolean, Integer, JSON, String
from stringcase import snakecase

from pyvoog.db import get_session
from pyvoog.exceptions import ValidationError
from pyvoog.util import Undefined
from pyvoog.validatable import Validatable

class SchemaGenerator:
    skipped_fields = ["id"]

    @classmethod
    def generate_schema(cls, model):

        """ Generate a skeleton Marshmallow schema by reflecting on an
        SQLAlchemy model. Include both columns and virtual attributes. Virtual
        attrs are `allow_none` as not to clash with field validation down the
        pipeline.
        """

        schema_dict = {}

        for c in inspect(model).c:
            schema_dict.update(cls._get_column_schema(c))

        for vattr in cls._get_vattrs(model):
            schema_dict[vattr.name] = fields.Field(allow_none=True)

        return Schema.from_dict(schema_dict)

    @classmethod
    def _get_column_schema(cls, c):
        _type = c.type
        column_schema = {}
        field_kwargs = {}

        if (c.name in cls.skipped_fields or isinstance(c, SchemalessColumn)):
            return column_schema
        elif c.nullable:
            field_kwargs['allow_none'] = True
        elif not c.nullable:
            field_kwargs['required'] = True

        if isinstance(_type, Boolean):
            column_schema[c.name] = fields.Boolean(**field_kwargs)
        elif isinstance(_type, Integer):
            column_schema[c.name] = fields.Integer(**field_kwargs)
        elif isinstance(_type, String):
            validator = None

            if _type.length is not None:
                validator = validate.Length(max=_type.length)

            column_schema[c.name] = fields.Str(validate=validator, **field_kwargs)
        else:
            column_schema[c.name] = fields.Field(**field_kwargs)

        return column_schema

    @classmethod
    def _get_vattrs(cls, model):
        return filter(
            lambda attr: isinstance(attr, VirtualAttribute),
            model.__dict__.values()
        )

class SchemalessColumn(Column):

    """ Represent a JSON database column for holding virtual attributes
    (unstructured data).
    """

    def __init__(self, **kwargs):
        super().__init__(MutableDict.as_mutable(JSON), **kwargs)

class VirtualAttribute(Validatable):

    """ Support virtual attributes, i.e. those not backed by a distinct
    database column on a model. These are routed to a JSON or TEXT field in
    the model's table. The attribute name is automatically deduced by
    ModelMetaClass.
    """

    ALLOWED_PLAIN_DEFAULT_TYPES = (
        bool,
        bytes,
        float,
        frozenset,
        int,
        str,
        tuple,
        types.NoneType,
        types.LambdaType,
        types.FunctionType,
    )

    def __init__(self, default=Undefined, schemaless_field="schemaless"):

        """ Attributes:

        - default - On lookup, return this value instead of raising a KeyError
          if the lookup fails. Note that this may be a callable, invoked
          whenever a fallback value is required. A callable must be used for
          referenced data structures (like dicts or lists) to avoid leakage
          across model instances.
        - schemaless_field - Name of the backing JSON or TEXT column.
        """

        super().__init__()

        if default != Undefined and not any(
            isinstance(default, t) for t in self.ALLOWED_PLAIN_DEFAULT_TYPES
        ):
            raise TypeError(
                f"It is not allowed to directly assign a VirtualAttribute default of type "
                f"`{type(default).__name__}`, as mutable types may leak data across "
                "model instances. Use the callable form of `default` instead."
            )

        self.default = default
        self.schemaless_field = schemaless_field

    @property
    def name(self):
        return self.attr_name

    def __get__(self, obj, objtype=None):
        try:
            return getattr(obj, self.schemaless_field)[self.attr_name]
        except (KeyError, TypeError):
            if self.default == Undefined:
                raise
            elif callable(self.default):
                return self.default()
            else:
                return self.default

    def __set__(self, obj, value):
        if obj.schemaless is None:
            obj.schemaless = {}

        obj.schemaless[self.attr_name] = value

    def _set_attr_name(self, attr_name):
        self.attr_name = attr_name

class UTCTimeStamp(sa_types.TypeDecorator):

    """ An SQLAlchemy type decorator for converting and storing all incoming
    timestamps as UTC and returning these as such. Note that SQLite drops
    all time zone information, hence always storing timestamps as UTC is a
    workaround to that issue. If an incoming datetime is naive (i.e.
    contains no time zone information), it is considered to represent time
    in the system timezone. See:

    https://docs.python.org/3/library/datetime.html#datetime.datetime.astimezone
    https://mike.depalatis.net/blog/sqlalchemy-timestamps.html
    """

    impl = sa_types.DateTime
    cache_ok = True

    def process_bind_param(self, value: datetime, dialect):
        return value.astimezone(timezone.utc)

    def process_result_value(self, value, dialect):
        if value.tzinfo is None:
            return value.replace(tzinfo=timezone.utc)

        return value.astimezone(timezone.utc)

class ModelMetaclass(DeclarativeMeta):

    """ Model base metaclass. """

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)

        cls._set_va_attr_names()
        cls._declare_timestamps()
        cls._attach_init_listener()

    def _set_va_attr_names(cls):

        """ Attach attribute names to VirtualAttributes. """

        for k, v in cls.__dict__.items():
            if k.startswith("_") or not isinstance(v, VirtualAttribute):
                continue

            v._set_attr_name(k)

    def _declare_timestamps(cls):
        if not getattr(cls, "include_timestamps", False):
            return

        cls.created_at = Column(UTCTimeStamp(), default=func.now())
        cls.updated_at = Column(UTCTimeStamp(), default=func.now(), onupdate=func.now())

    def _attach_init_listener(cls):

        """ When a model is about to be (explicitly) instantiated, init it with
        the default scope (if any). See also:

        https://docs.sqlalchemy.org/en/20/orm/events.html#sqlalchemy.orm.InstanceEvents.init
        """

        if hasattr(cls, "default_scope"):
            listen(cls, "init", cls._apply_default_scope)

class Model:

    """ A model base class providing the following facilities:

    - The table name is automatically deduced from the class name.
    - A Marshmallow schema is generated based on the model schema. The
      `validate` method checks the object against the schema and any
      constraints installed via Validatable. All errors are merged into a
      single data structure and signalled as a ValidationError. Note that
      ValidatingSession also automatically runs validations before flush on
      new and dirty session members.
    - Default scopes. If a model class has the `default_scope` attribute
      defined, it is expected to be a callable returning a dict of keyword
      attributes to pass to SQLAlchemy's `filter_by`. A statement with the
      scope applied can be retrieved via the `get_scoped_query` method or
      the `scoped_query` property.

    Model attributes may either have a 1:1 mapping to database columns
    (`sqlalchemy.Column` or its subclasses) or be VirtualAttributes which
    are stored in the `schemaless` JSON field. Validations may be attached
    to either.

    TODO: Override default constructor to allow passing vattrs at
    instantiation. See:

    https://docs.sqlalchemy.org/en/14/orm/mapping_styles.html#default-constructor
    """

    id = Column(Integer, primary_key=True)

    @declared_attr
    def __tablename__(cls):
        return snakecase(cls.__name__)

    @classmethod
    def __declare_last__(cls):
        cls.__schema__ = SchemaGenerator.generate_schema(cls)

    @classmethod
    def get_unscoped_query(cls, *args):

        """ Return an unscoped Select. If any arguments are passed, these are
        forwarded to `select` and the table is explicitly specified via
        `select_from`.
        """

        return select(*args).select_from(cls) if args else select(cls)

    @classmethod
    def get_query(cls, *args):

        """ Return a Select with any default scope defined by the model
        applied. Arguments are forwarded to `get_unscoped_query`.
        """

        return cls._apply_default_scope_to_stmt(cls.get_unscoped_query(*args))

    @classmethod
    def get_unscoped_statement(cls, verb, *args):

        """ A generalization of `get_unscoped_query` for any SQL verb. Return an
        unscoped statement object. If any arguments are passed, these are
        forwarded to the selected method.
        """

        return getattr(sqlalchemy, verb)(cls, *args)

    @classmethod
    def get_statement(cls, verb, *args):

        """ A generalization of `get_query` for any SQL verb. The default scope
        has been applied to the returned statement.
        """

        return cls._apply_default_scope_to_stmt(cls.get_unscoped_statement(verb, *args))

    def validate(self):
        schema = self.__class__.__schema__()
        attrs = self._get_attr_dict()
        errors = schema.validate(attrs)

        for (attr_name, messages) in self._run_attr_validations():
            always_merger.merge(errors, {attr_name: messages})

        if errors:
            raise ValidationError(errors, None, attrs)

    def save(self):

        """ A convenience method wrapping determining the appropriate session
        (any session the object is already attached to, or the per-request
        session), adding the object to the session and committing.
        """

        session = object_session(self) or get_session()

        try:
            session.add(self)
            session.commit()
        except Exception:
            session.rollback()
            raise

    def as_dict(self):
        return {"id": self.id, **self._get_attr_dict()}

    @property
    def attributes(self):
        raise NotImplementedError()

    @attributes.setter
    def attributes(self, attrs):
        for k, v in attrs.items():
            setattr(self, k, v)

    @classmethod
    def _apply_default_scope_to_stmt(cls, stmt):
        scope = getattr(cls, "default_scope", None)

        if scope:
            stmt = stmt.filter_by(**scope())

        return stmt

    def _run_attr_validations(self):
        columns = inspect(self.__class__).c
        vattrs = self._get_vattrs()

        for c in itertools.chain(columns, vattrs):
            if isinstance(c, Validatable):
                try:
                    c.is_valid(self)
                except ValidationError as e:
                    yield (c.name, e.messages)

    def _get_vattrs(self):
        return filter(
            lambda attr: isinstance(attr, VirtualAttribute),
            self.__class__.__dict__.values()
        )

    def _get_attr_dict(self):
        schema_fields = self.__class__.__schema__().fields
        return {k: getattr(self, k) for k, v in schema_fields.items()}

    def _apply_default_scope(self, args, kwargs):
        scope = getattr(self, "default_scope", None)

        if scope:
            kwargs |= scope.__func__()

Model = declarative_base(cls=Model, metaclass=ModelMetaclass)



from deepmerge import merge_or_raise
from sqlalchemy import Column

from pyvoog.exceptions import ValidationError
from pyvoog.validations import Custom

class Validatable:

    """ A mixin class providing the `validate` method for installing
    validations on a class and the `is_valid` method for running these.

    A validator is a class:

    - whose constructor conforms to the signature (self, field, ...) where
      `field` is either a `ValidatingColumn` or another Validatable, and
    - which provides a `run` method with the signature (self, obj) where
      `obj` is an instance of an SQLAlchemy model. The method is invoked to
      check the installed constraint against a concrete object. If the
      validation fails, a `ValidationError` must be raised with a string,
      list of error messages, or a dict.

    Note that some validators only accept a subclass of `sqlalchemy.Column`
    as a validatable; in this case, instantiating the validator fails with a
    generic Validatable.

    As a convenience, a string may be passed as the validator. This is a
    shortcut to using the Custom validator; the string is passed as
    `validator_name` to Custom on instantiation.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._validations = []

    def validate(self, Validation, **kwargs):
        if type(Validation) is str:
            kwargs = dict(validator_name=Validation)
            Validation = Custom

        self._validations.append(Validation(self, **kwargs))
        return self

    def is_valid(self, obj):

        """ Run all validations on an object, merge the payloads of any
        ValidationErrors and raise these as a combined ValidationError.
        """

        messages = None

        for validation in self._validations:
            try:
                validation.run(obj)
            except ValidationError as e:
                messages = merge_or_raise.merge(messages, e.messages) if messages else e.messages

        if messages:
            raise ValidationError(messages)

class ValidatingColumn(Validatable, Column):

    """ A convenience class mixing in Validatable with Column. """

    inherit_cache = True
